# SPDX-License-Identifier: MIT

from __future__ import annotations

from base64 import b64decode
from typing import Iterator

from ..track import Track

__all__ = ("decode_track",)


def _bytes_to_int(bytes: list[int]) -> int:
    result = 0
    for b in bytes:
        result = result * 256 + b

    return result


class _TrackDataIterator(Iterator[int]):
    __slots__ = ("__previous_null", "flags", "__iterable", "size", "version")

    _TRACK_INFO_VERSIONED = 1
    _FLAG_MASK = 0xC0000000

    def __init__(self, data: bytes):
        self.__iterable = iter(data)

        # https://github.com/sedmelluq/lavaplayer/blob/97a8efecfe3cb79da4d7d0422de0179e18c30947/main/src/main/java/com/sedmelluq/discord/lavaplayer/tools/io/MessageInput.java#L37
        value = self.read_int(bit_length=4)
        self.flags = (value & self._FLAG_MASK) >> 30
        self.size = value & 0x3FFFFFFF
        self.version = (
            self.read_int(bit_length=1) & 0xFF
            if (self.flags & self._TRACK_INFO_VERSIONED) != 0
            else 1
        )
        self.__previous_null = False
        """A value which only bool cares about, string sets it when it leaves."""

    def __next__(self) -> int:
        b = next(self.__iterable)

        try:
            if self.__previous_null and b:
                self.__previous_null = False
        # Somehow this can have an AttributeError on creation?
        except AttributeError:
            pass

        return b

    def read_line(self) -> list[int]:
        line: list[int] = []

        # Checks against NULL bytes - end of "line".
        # Check "or not line" to skip multiple prefixed NULL bytes.
        while byte := next(self) or not line:
            # ASCII for header beginning.
            if byte <= 0x20 and not line:
                continue

            line.append(byte)

        self.__previous_null = True
        return line

    def read_str(self) -> str:
        # Support all unicode code points.
        return "".join(map(chr, self.read_line()))

    def read_int(self, *, bit_length: int = 8) -> int:
        return _bytes_to_int([next(self) for _ in range(bit_length)])

    def read_bool(self) -> bool:
        if self.__previous_null:
            self.__previous_null = False
            return False

        return bool(next(self))

    def read_nullable_str(self) -> str | None:
        # URLs seem to be prefixed with `+`
        exists = self.read_bool()

        if not exists:
            return None

        if (byte := next(self)) != 0:
            raise ValueError(
                "Attempted to traverse a track id "
                f"and came across an unexpected character: {byte}."
            )

        return self.read_str()


def decode_track(track_id: str) -> Track:
    """Decode a track id into a Track object.

    .. warning::

        This still needs a lot of testing and may not work for many tracks.
        Custom fields on plugins are not implemented and positions in tracks
        may not work. Use at your own risk.

    Parameters
    ----------
    track_id:
        The track id to decode.

    Returns
    -------
    Track:
        The decoded track.
    """

    raw = b64decode(track_id)
    print(raw)
    iterator = _TrackDataIterator(raw)

    # https://github.com/sedmelluq/lavaplayer/blob/97a8efecfe3cb79da4d7d0422de0179e18c30947/main/src/main/java/com/sedmelluq/discord/lavaplayer/player/DefaultAudioPlayerManager.java#L268
    title = iterator.read_str()
    author = iterator.read_str()
    length = iterator.read_int()
    identifier = iterator.read_str()
    stream = iterator.read_bool()
    if iterator.version >= 2:
        uri = iterator.read_nullable_str()
        if uri is not None:
            while not uri.startswith("http"):
                uri = uri[1:]
    else:
        uri = None

    source = iterator.read_str()

    # HACK: Needs testing with an encoded track with a position.
    # Who knows how you get that!?!?!?!?!??!?!?
    try:
        position = iterator.read_int()
    except StopIteration:
        position = 0

    # TODO: remove once tested enough
    # print("title:".ljust(20), title)
    # print("author:".ljust(20), author)
    # print("length:".ljust(20), length)
    # print("identifier:".ljust(20), identifier)
    # print("stream:".ljust(20), stream)
    # print("uri:".ljust(20), uri)
    # print("source:".ljust(20), source)
    # print("position:".ljust(20), position)

    return Track(
        track_id=track_id,
        title=title,
        author=author,
        length=length,
        identifier=identifier,
        stream=stream,
        uri=uri,
        source=source,
        # https://github.com/sedmelluq/lavaplayer/blob/84f0d36a6b32a40bf9ab290ca5590d578ddc5d24/main/src/main/java/com/sedmelluq/discord/lavaplayer/track/BaseAudioTrack.java#L71
        seekable=not stream,
        position=position,
    )



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Callable, Generic, TypeVar

T = TypeVar("T")

__all__ = ("classproperty",)


class _ClassPropertyDescriptor(Generic[T]):
    """A descriptor that mimics the behavior of a property, but for classmethods."""

    def __init__(self, fget: classmethod[T] | staticmethod[T], fset: None = None):
        self.fget = fget

    def __get__(self, instance: object, owner: type | None = None) -> T:
        if owner is None:
            owner = type(instance)
        return self.fget.__get__(instance, owner)()


def classproperty(
    func: Callable[..., T] | classmethod[T] | staticmethod[T]
) -> _ClassPropertyDescriptor[T]:
    """A decorator that mimics the behavior of a property, but for classmethods.

    Parameters
    ----------
    func:
        The function to decorate.
    """

    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return _ClassPropertyDescriptor(func)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import re
from enum import Enum, auto
from logging import getLogger
from random import choice
from typing import Callable, List, Optional

from .node import Node
from .type_variables import ClientT

StrategyCallable = Callable[
    [List[Node[ClientT]], int, Optional[int], Optional[str]], List[Node[ClientT]]
]
"""Represents a strategy callable.

This accepts ``nodes``, ``guild_id``, ``shard_count`` and ``endpoint`` and returns
a list of nodes.
"""

_log = getLogger(__name__)
__all__ = ("Strategy", "call_strategy")


class Strategy(Enum):
    """Represents a strategy for selecting a node."""

    LOCATION = auto()
    """Selects a node based on the region the guild is in."""

    RANDOM = auto()
    """Selects a random node."""

    SHARD = auto()
    """Selects a node based on the shard ID of the guild."""

    USAGE = auto()
    """Selects a node based on the least used node."""


def shard_strategy(
    nodes: list[Node[ClientT]], guild_id: int, shard_count: int | None, _: str | None
) -> list[Node[ClientT]]:
    """Selects a node based on the shard ID of the guild.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    guild_id:
        The ID of the guild to select a node for.
    shard_count:
        The number of shards the bot is using.
    _:
        Unused parameter.
    """

    if shard_count is None:
        shard_count = 1

    shard_id = (guild_id >> 22) % shard_count

    return list(
        filter(lambda node: node.shard_ids is None or shard_id in node.shard_ids, nodes)
    )


_REGION_REGEX = re.compile(r"(?:vip-)?(?P<region>[a-z-]{1,15})\d{1,5}\.discord\.media")


def location_strategy(
    nodes: list[Node[ClientT]], _: int, __: int | None, endpoint: str | None
) -> list[Node[ClientT]]:
    """Selects a node based on the region the guild is in.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    _:
        Unused parameter.
    __:
        Unused parameter.
    endpoint:
        The endpoint of the guild to select a node for.
    """

    if endpoint is None:
        return nodes

    match = _REGION_REGEX.match(endpoint)

    if not match:
        _log.error("Failed to find the region in an endpoint, defaulting to all nodes.")
        return nodes

    voice_region = match.group("region")

    if not voice_region:
        _log.error(
            "Failed to match endpoint %s (match: %s) to a region, "
            "defaulting to all nodes.",
            endpoint,
            match.group("region"),
        )
        return nodes

    regional_nodes = list(
        filter(
            lambda node: node.regions is not None and voice_region in node.regions,
            nodes,
        )
    )

    if not regional_nodes:
        _log.error("No nodes found for the region, defaulting to all nodes.")
        return nodes

    return regional_nodes


def usage_strategy(
    nodes: list[Node[ClientT]], _: int, __: int | None, ___: str | None
) -> list[Node[ClientT]]:
    """Selects a node based on the least used node.

    This is calculated using :attr:`Node.weight`.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    _:
        Unused parameter.
    __:
        Unused parameter.
    ___:
        Unused parameter.
    """

    # max() would be nice, however if all nodes have no stats, it returns the first.

    lowest = None

    for node in nodes:
        weight = node.weight

        if lowest is None or weight < lowest:
            lowest = weight

    if lowest is None:
        return nodes

    return list(filter(lambda node: node.weight == lowest, nodes))


def random_strategy(
    nodes: list[Node[ClientT]], _: int, __: int | None, ___: str | None
) -> list[Node[ClientT]]:
    """Selects a random node.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    _:
        Unused parameter.
    __:
        Unused parameter.
    ___:
        Unused parameter.
    """

    return [choice(nodes)]


# Used for ClientT preservation.
def call_strategy(
    strategy: Strategy,
    nodes: list[Node[ClientT]],
    guild_id: int,
    shard_count: int | None,
    endpoint: str | None,
) -> list[Node[ClientT]]:
    """Calls a strategy.

    Parameters
    ----------
    strategy:
        The strategy to call.
    nodes:
        The nodes to select from.
    guild_id:
        The ID of the guild to select a node for.
    shard_count:
        The number of shards the bot is using.
    endpoint:
        The endpoint of the voice client to select a node for.
    """

    if strategy is Strategy.LOCATION:
        return location_strategy(nodes, guild_id, shard_count, endpoint)
    elif strategy is Strategy.RANDOM:
        return random_strategy(nodes, guild_id, shard_count, endpoint)
    elif strategy is Strategy.SHARD:
        return shard_strategy(nodes, guild_id, shard_count, endpoint)
    elif strategy is Strategy.USAGE:
        return usage_strategy(nodes, guild_id, shard_count, endpoint)
    else:
        raise ValueError(f"Unknown strategy {strategy}")



# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Sequence
from functools import partial
from logging import getLogger
from random import choice
from typing import TYPE_CHECKING, Any, Generic, TypeVar, Union

from .errors import NoNodesAvailable
from .node import Node
from .strategy import Strategy, StrategyCallable, call_strategy
from .type_variables import ClientT
from .utils import classproperty

if TYPE_CHECKING:
    from typing import ClassVar

    import aiohttp

    from .region import Group, Region, VoiceRegion


T = TypeVar("T")
__all__ = ("NodePool",)

_log = getLogger(__name__)


StrategyList = Union[
    StrategyCallable[ClientT],
    Sequence[Strategy],
    Sequence[StrategyCallable[ClientT]],
    Sequence[Union[Strategy, StrategyCallable[ClientT]]],
]
"""A type hint for a list of strategies to select the best node.

This can either be:

- A single :data:`.StrategyCallable`.
- A sequence of :class:`.Strategy`\\s.
- A sequence of :data:`.StrategyCallable`\\s.
- A sequence of :class:`.Strategy`\\s and :data:`.StrategyCallable`\\s.
"""


class NodePool(Generic[ClientT]):
    """A class that manages nodes and chooses them based on strategies.

    Parameters
    ----------
    client:
        The client to use to connect to the nodes.
    default_strategies:
        The default strategies to use when selecting a node. Defaults to
        [:attr:`Strategy.SHARD`, :attr:`Strategy.LOCATION`, :attr:`Strategy.USAGE`].
    """

    __slots__ = ()

    # Generics as expected, do not work in class variables.
    # Don't fear, the public methods using this are typed well.
    _nodes: ClassVar[dict[str, Node[Any]]] = {}
    _node_regions: ClassVar[dict[VoiceRegion, set[Node[Any]]]] = {}
    _node_shards: ClassVar[dict[int, set[Node[Any]]]] = {}
    _client: ClientT | None = None

    def __init__(
        self,
        client: ClientT,
        default_strategies: StrategyList[ClientT] | None = None,
    ) -> None:
        NodePool._client = client

        NodePool._default_strategies = (
            [
                Strategy.SHARD,
                Strategy.LOCATION,
                Strategy.USAGE,
            ]
            if default_strategies is None
            else default_strategies
        )

    @classproperty
    def label_to_node(cls) -> dict[str, Node[ClientT]]:
        """A dictionary mapping node labels to nodes."""

        return cls._nodes

    @classproperty
    def nodes(cls) -> list[Node[ClientT]]:
        """A list of all nodes."""

        return list(cls._nodes.values())

    async def create_node(
        self,
        *,
        host: str,
        port: int,
        label: str,
        password: str,
        secure: bool = False,
        heartbeat: int = 30,
        timeout: float = 10,
        session: aiohttp.ClientSession | None = None,
        resume_key: str | None = None,
        regions: Sequence[Group | Region | VoiceRegion] | None = None,
        shard_ids: Sequence[int] | None = None,
    ) -> Node[ClientT]:
        """Create a node and connect it.

        The parameters here relate to :class:`Node`.

        Parameters
        ----------
        host:
            The host of the node to connect to.
        port:
            The port of the node to connect to.
        label:
            The label of the node used to identify it.
        password:
            The password of the node to authenticate.
        secure:
            Whether to use SSL (TLS) or not.
        heartbeat:
            The interval to send heartbeats to the node websocket connection.
        timeout:
            The timeout to use for the node websocket connection.
        session: :data:`~typing.Optional`\\ [:class:`aiohttp.ClientSession`]
            The session to use for the node websocket connection.
        resume_key:
            The key to use to resume the node websocket connection.
            This is used to resume a connection if the connection is lost.
            Defaults to using the `host`, `port` and `label`.
        regions:
            The regions this node is best to connect to.
            This is used to determine if this is the best node for a connection
            with :attr:`Strategy.LOCATION`.
        shard_ids:
            The shard IDs this node is best to connect to.
            This is used to determine if this is the best node for a connection
            with :attr:`Strategy.SHARD`.

        Returns
        -------
        :class:`Node`
            The created node.

        Raises
        ------
        RuntimeError
            If the node pool has not been initialized.
        """

        if self._client is None:
            raise RuntimeError("NodePool has not been initialized.")

        node = Node(
            host=host,
            port=port,
            label=label,
            password=password,
            client=self._client,
            secure=secure,
            heartbeat=heartbeat,
            timeout=timeout,
            session=session,
            resume_key=resume_key,
            regions=regions,
            shard_ids=shard_ids,
        )

        self._nodes[label] = node

        # Add to dictionaries, creating a set or extending it if needed.
        if node.regions:
            for region in node.regions:
                self._node_regions[region] = {
                    node,
                    *self._node_regions.get(region, set()),
                }

        if node.shard_ids:
            for shard_id in node.shard_ids:
                self._node_shards[shard_id] = {
                    node,
                    *self._node_shards.get(shard_id, set()),
                }

        _log.info("Created node, connecting it...", extra={"label": label})
        await node.connect()

        return node

    @classmethod
    def get_node(
        cls,
        *,
        guild_id: str | int,
        endpoint: str | None,
        strategies: StrategyList[ClientT] | None = None,
    ) -> Node[ClientT]:
        """Get a node based on the given strategies.

        Parameters
        ----------
        guild_id:
            The guild ID to get a node for.
        endpoint:
            The endpoint to get a node for.
        strategies:
            The strategies to use to get a node. If not provided, the default
            strategies will be used.

        Returns
        -------
        :class:`Node`
            The node to use.

        Raises
        ------
        RuntimeError
            If the node pool has not been initialized.
        """

        if cls._client is None:
            raise RuntimeError("NodePool has not been initialized.")

        actual_strategies: Sequence[StrategyCallable[ClientT] | Strategy]

        strategies = strategies or cls._default_strategies

        if callable(strategies):
            actual_strategies = [strategies]
        else:
            actual_strategies = strategies

        nodes = cls.nodes

        for strategy in actual_strategies:
            if isinstance(strategy, Strategy):
                strategy_callable = partial(call_strategy, strategy)
            else:
                strategy_callable = strategy

            nodes = strategy_callable(
                nodes, int(guild_id), cls._client.shard_count, endpoint
            )

            _log.debug(
                "Strategy %s returned nodes %s.",
                strategy.__name__ if callable(strategy) else strategy.name,
                ", ".join(n.label for n in nodes),
            )

            if len(nodes) == 1:
                return nodes[0]
            elif len(nodes) == 0:
                raise NoNodesAvailable

        return choice(nodes)

    @classmethod
    def get_random_node(cls) -> Node[ClientT]:
        """Get a random node.

        Returns
        -------
        :class:`Node`
            The random node.

        Raises
        ------
        ValueError
            If there are no nodes.
        """

        return choice(list(cls._nodes.values()))
